from collections import OrderedDict

from typing import Set, List, Optional, Union

from irrd.conf import PASSWORD_HASH_DUMMY_VALUE
from irrd.utils.pgp import get_gpg_instance
from .config import PASSWORD_HASHERS
from .fields import (RPSLTextField, RPSLIPv4PrefixField, RPSLIPv4PrefixesField, RPSLIPv6PrefixField,
                     RPSLIPv6PrefixesField, RPSLIPv4AddressRangeField, RPSLASNumberField, RPSLASBlockField,
                     RPSLSetNameField, RPSLEmailField, RPSLDNSNameField, RPSLGenericNameField, RPSLReferenceField,
                     RPSLReferenceListField, RPSLAuthField, RPSLRouteSetMembersField, RPSLChangedField)
from .parser import RPSLObject, UnknownRPSLObjectClassException

RPSL_ROUTE_OBJECT_CLASS_FOR_IP_VERSION = {
    4: 'route',
    6: 'route6',
}


def rpsl_object_from_text(text, strict_validation=True, default_source: Optional[str]=None) -> RPSLObject:
    rpsl_object_class = text.split(':', maxsplit=1)[0].strip()
    try:
        klass = OBJECT_CLASS_MAPPING[rpsl_object_class]
    except KeyError:
        raise UnknownRPSLObjectClassException(f'unknown object class: {rpsl_object_class}',
                                              rpsl_object_class=rpsl_object_class)
    return klass(from_text=text, strict_validation=strict_validation, default_source=default_source)


class RPSLAsBlock(RPSLObject):
    fields = OrderedDict([
        ('as-block', RPSLASBlockField(primary_key=True, lookup_key=True)),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, multiple=True, referring=['role', 'person'])),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLAsSet(RPSLObject):
    fields = OrderedDict([
        ('as-set', RPSLSetNameField(primary_key=True, lookup_key=True, prefix='AS')),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('members', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['aut-num', 'as-set'], strong=False)),
        ('mbrs-by-ref', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['mntner'], allow_kw_any=True, strong=False)),
        ('admin-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLAutNum(RPSLObject):
    fields = OrderedDict([
        ('aut-num', RPSLASNumberField(primary_key=True, lookup_key=True)),
        ('as-name', RPSLGenericNameField(allowed_prefixes=['AS'])),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('member-of', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['as-set'], strong=False)),
        ('import', RPSLTextField(optional=True, multiple=True)),
        ('mp-import', RPSLTextField(optional=True, multiple=True)),
        ('import-via', RPSLTextField(optional=True, multiple=True)),
        ('export', RPSLTextField(optional=True, multiple=True)),
        ('mp-export', RPSLTextField(optional=True, multiple=True)),
        ('export-via', RPSLTextField(optional=True, multiple=True)),
        ('default', RPSLTextField(optional=True, multiple=True)),
        ('mp-default', RPSLTextField(optional=True, multiple=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, multiple=True, referring=['role', 'person'])),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLDomain(RPSLObject):
    fields = OrderedDict([
        ('domain', RPSLTextField(primary_key=True, lookup_key=True)),  # reverse delegation address (range), v4/v6/enum
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, multiple=True, referring=['role', 'person'])),
        ('zone-c', RPSLReferenceField(lookup_key=True, multiple=True, referring=['role', 'person'])),
        ('nserver', RPSLTextField(optional=True, multiple=True)),  # DNS name, possibly followed v4/v6
        ('sub-dom', RPSLTextField(optional=True, multiple=True)),
        ('dom-net', RPSLTextField(optional=True, multiple=True)),
        ('refer', RPSLTextField(optional=True)),  # ???
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLFilterSet(RPSLObject):
    fields = OrderedDict([
        ('filter-set', RPSLSetNameField(primary_key=True, lookup_key=True, prefix='FLTR')),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('filter', RPSLTextField()),
        ('mp-filter', RPSLTextField(optional=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLInetRtr(RPSLObject):
    fields = OrderedDict([
        ('inet-rtr', RPSLDNSNameField(primary_key=True, lookup_key=True)),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('alias', RPSLDNSNameField(optional=True, multiple=True)),
        ('local-as', RPSLASNumberField()),
        ('ifaddr', RPSLTextField(optional=True, multiple=True)),
        ('interface', RPSLTextField(optional=True, multiple=True)),
        ('peer', RPSLTextField(optional=True, multiple=True)),
        ('mp-peer', RPSLTextField(optional=True, multiple=True)),
        ('member-of', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['rtr-set'], strong=False)),
        ('rs-in', RPSLTextField(optional=True)),
        ('rs-out', RPSLTextField(optional=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLInet6Num(RPSLObject):
    fields = OrderedDict([
        ('inet6num', RPSLIPv6PrefixField(primary_key=True, lookup_key=True)),
        ('netname', RPSLTextField()),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('country', RPSLTextField(multiple=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, multiple=True, referring=['role', 'person'])),
        ('rev-srv', RPSLTextField(optional=True, multiple=True)),
        ('status', RPSLTextField()),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLInetnum(RPSLObject):
    fields = OrderedDict([
        ('inetnum', RPSLIPv4AddressRangeField(primary_key=True, lookup_key=True)),
        ('netname', RPSLTextField()),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('country', RPSLTextField(multiple=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, multiple=True, referring=['role', 'person'])),
        ('rev-srv', RPSLTextField(optional=True, multiple=True)),
        ('status', RPSLTextField()),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLKeyCert(RPSLObject):
    fields = OrderedDict([
        ('key-cert', RPSLGenericNameField(primary_key=True, lookup_key=True)),
        ('method', RPSLTextField(optional=True)),  # Fixed to PGP
        ('owner', RPSLTextField(optional=True, multiple=True)),  # key owner, autogenerate
        ('fingerpr', RPSLTextField(optional=True)),  # fingerprint, autogenerate
        ('certif', RPSLTextField(multiple=True)),  # Actual key
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])

    def clean(self) -> bool:
        """
        Validate the PGP key and update relevant attributes.

        In key-cert objects, the method, owner and fingerpr objects should be
        auto-generated based on the certif object. The certif object should be
        a valid PGP key, matching the ID in the primary key.

        Note that the PGP key is imported into the keyring every time this is
        called - this is intentional, to decouple the RPSL database state from
        the gpg keyring state.
        """
        if not super().clean():
            return False  # pragma: no cover

        gpg = get_gpg_instance()
        certif_data = '\n'.join(self.parsed_data.get('certif', [])).replace(',', '\n')
        result = gpg.import_keys(certif_data)

        if len(result.fingerprints) != 1:
            msg = 'Unable to read public PGP key: key corrupt or multiple keys provided'
            if result.results:
                msg = f'{msg}: {result.results[0]["text"]}'
            self.messages.error(msg)
            return False

        self.fingerprint = result.fingerprints[0]
        expected_object_name = 'PGPKEY-' + self.fingerprint[-8:]
        actual_object_name = self.parsed_data['key-cert'].upper()
        fingerprint_formatted = self.format_fingerprint(self.fingerprint)

        if expected_object_name != actual_object_name:
            self.messages.error(
                f'Invalid object name {actual_object_name}: does not match key fingerprint {fingerprint_formatted}, '
                f'expected object name {expected_object_name}'
            )
            return False

        self._update_attribute_value('fingerpr', fingerprint_formatted)
        self._update_attribute_value('owner', gpg.list_keys(keys=self.fingerprint)[0]['uids'])
        self._update_attribute_value('method', 'PGP')

        return True

    # This API is correct, but not very practical.
    # In typical cases, the PGP key used to sign a message is not known until
    # the PGP signature is actually parsed. More useful is a generic method to find
    # which key signed a message, which can then be stored and compared to key-cert's later.
    # This method will probably be extracted to the update handler.
    def verify(self, message: str) -> bool:
        gpg = get_gpg_instance()
        result = gpg.verify(message)
        return result.valid and result.key_status is None and \
            self.format_fingerprint(result.fingerprint) == self.parsed_data['fingerpr']

    @staticmethod
    def format_fingerprint(fingerprint: str) -> str:
        """Format a PGP fingerprint into sections of 4 characters, separated by spaces."""
        string_parts = []
        for idx in range(0, 40, 4):
            string_parts.append(fingerprint[idx:idx + 4])
            if idx == 16:
                string_parts.append('')
        return ' '.join(string_parts)


class RPSLMntner(RPSLObject):
    fields = OrderedDict([
        ('mntner', RPSLGenericNameField(primary_key=True, lookup_key=True)),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('upd-to', RPSLEmailField(multiple=True)),
        ('mnt-nfy', RPSLEmailField(optional=True, multiple=True)),
        ('auth', RPSLAuthField(multiple=True)),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])

    def clean(self):
        """Check whether either all hash values are dummy hashes, or none."""
        if not super().clean():
            return False  # pragma: no cover

        dummy_matches = [auth[1] == PASSWORD_HASH_DUMMY_VALUE for auth in self._auth_lines(True)]
        if any(dummy_matches) and not all(dummy_matches):
            self.messages.error('Either all password auth hashes in a submitted mntner must be dummy objects, or none.')

    def verify_auth(self, passwords: List[str], keycert_obj_pk: Optional[str]=None) -> bool:
        """
        Verify whether one of a given list of passwords matches
        any of the auth hashes in this object, or match the
        keycert object PK.
        """
        for auth in self.parsed_data.get('auth', []):
            if keycert_obj_pk and auth.upper() == keycert_obj_pk.upper():
                return True
            if ' ' not in auth:
                continue
            scheme, hash = auth.split(' ', 1)
            hasher = PASSWORD_HASHERS.get(scheme.upper())
            if hasher:
                for password in passwords:
                    try:
                        if hasher.verify(password, hash):
                            return True
                    except ValueError:
                        pass
        return False

    def has_dummy_auth_value(self) -> bool:
        """
        Check whether this object has dummy auth hashes.
        If clean() has returned successfully before, the answer from this method
        means that either all or no hashes have dummy values.
        """
        auth_values = [auth[1] for auth in self._auth_lines(password_hashes=True)]
        return bool(auth_values) and all([value == PASSWORD_HASH_DUMMY_VALUE for value in auth_values])

    def force_single_new_password(self, password) -> None:
        """
        Overwrite all auth hashes with a single new hash for the provided password.
        Retains other methods, i.e. PGPKEY.
        """
        hash = 'MD5-PW ' + PASSWORD_HASHERS['MD5-PW'].hash(password)
        auths = self._auth_lines(password_hashes=False)
        auths.append(hash)
        self._update_attribute_value('auth', auths)

    def _auth_lines(self, password_hashes=True) -> List[Union[str, List[str]]]:
        """
        Return a list of auth values in this object.
        If password_hashes=False, returns only non-hash (i.e. PGPKEY) lines.
        If password_hashes=True, returns a list of lists, each inner list containing
        the hash method and the hash.
        """
        lines = self.parsed_data.get('auth', [])
        if password_hashes is True:
            return [auth.split(' ', 1) for auth in lines if ' ' in auth]
        return [auth for auth in lines if ' ' not in auth]


class RPSLPeeringSet(RPSLObject):
    fields = OrderedDict([
        ('peering-set', RPSLSetNameField(primary_key=True, lookup_key=True, prefix='PRNG')),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('peering', RPSLTextField(optional=True, multiple=True)),
        ('mp-peering', RPSLTextField(optional=True, multiple=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLPerson(RPSLObject):
    fields = OrderedDict([
        ('person', RPSLTextField(lookup_key=True)),
        ('address', RPSLTextField(multiple=True)),
        ('phone', RPSLTextField(multiple=True)),
        ('fax-no', RPSLTextField(optional=True, multiple=True)),
        ('e-mail', RPSLEmailField(multiple=True)),
        ('nic-hdl', RPSLGenericNameField(primary_key=True, lookup_key=True, non_strict_allow_any=True)),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLRole(RPSLObject):
    fields = OrderedDict([
        ('role', RPSLTextField(lookup_key=True)),
        ('trouble', RPSLTextField(optional=True, multiple=True)),
        ('address', RPSLTextField(multiple=True)),
        ('phone', RPSLTextField(multiple=True)),
        ('fax-no', RPSLTextField(optional=True, multiple=True)),
        ('e-mail', RPSLEmailField(multiple=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('nic-hdl', RPSLGenericNameField(primary_key=True, lookup_key=True, non_strict_allow_any=True)),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLRoute(RPSLObject):
    rpki_relevant = True
    discarded_fields = ['rpki-ov-state']
    fields = OrderedDict([
        ('route', RPSLIPv4PrefixField(primary_key=True, lookup_key=True)),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('origin', RPSLASNumberField(primary_key=True)),
        ('holes', RPSLIPv4PrefixesField(optional=True, multiple=True)),
        ('member-of', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['route-set'], strong=False)),
        ('inject', RPSLTextField(optional=True, multiple=True)),
        ('aggr-bndry', RPSLTextField(optional=True)),
        ('aggr-mtd', RPSLTextField(optional=True)),
        ('export-comps', RPSLTextField(optional=True)),
        ('components', RPSLTextField(optional=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('geoidx', RPSLTextField(optional=True, multiple=True)),
        ('roa-uri', RPSLTextField(optional=True)),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLRouteSet(RPSLObject):
    fields = OrderedDict([
        ('route-set', RPSLSetNameField(primary_key=True, lookup_key=True, prefix='RS')),
        ('members', RPSLRouteSetMembersField(ip_version=4, lookup_key=True, optional=True, multiple=True)),
        ('mp-members', RPSLRouteSetMembersField(ip_version=None, lookup_key=True, optional=True, multiple=True)),
        ('mbrs-by-ref', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['mntner'], allow_kw_any=True, strong=False)),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLRoute6(RPSLObject):
    rpki_relevant = True
    discarded_fields = ['rpki-ov-state']
    fields = OrderedDict([
        ('route6', RPSLIPv6PrefixField(primary_key=True, lookup_key=True)),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('origin', RPSLASNumberField(primary_key=True)),
        ('holes', RPSLIPv6PrefixesField(optional=True, multiple=True)),
        ('member-of', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['route-set'], strong=False)),
        ('inject', RPSLTextField(optional=True, multiple=True)),
        ('aggr-bndry', RPSLTextField(optional=True)),
        ('aggr-mtd', RPSLTextField(optional=True)),
        ('export-comps', RPSLTextField(optional=True)),
        ('components', RPSLTextField(optional=True)),
        ('admin-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('geoidx', RPSLTextField(optional=True, multiple=True)),
        ('roa-uri', RPSLTextField(optional=True)),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


class RPSLRtrSet(RPSLObject):
    fields = OrderedDict([
        ('rtr-set', RPSLSetNameField(primary_key=True, lookup_key=True, prefix='RTRS')),
        ('descr', RPSLTextField(multiple=True, optional=True)),
        ('members', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['inet-rtr', 'rtr-set'], strong=False)),
        ('mp-members', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['inet-rtr', 'rtr-set'], strong=False)),
        ('mbrs-by-ref', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['mntner'], allow_kw_any=True, strong=False)),
        ('admin-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('tech-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
        ('remarks', RPSLTextField(optional=True, multiple=True)),
        ('notify', RPSLEmailField(optional=True, multiple=True)),
        ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
        ('changed', RPSLChangedField(optional=True, multiple=True)),
        ('source', RPSLGenericNameField()),
    ])


OBJECT_CLASS_MAPPING = {
    'as-block': RPSLAsBlock,
    'as-set': RPSLAsSet,
    'aut-num': RPSLAutNum,
    'domain': RPSLDomain,
    'filter-set': RPSLFilterSet,
    'inet-rtr': RPSLInetRtr,
    'inet6num': RPSLInet6Num,
    'inetnum': RPSLInetnum,
    'key-cert': RPSLKeyCert,
    'mntner': RPSLMntner,
    'peering-set': RPSLPeeringSet,
    'person': RPSLPerson,
    'role': RPSLRole,
    'route': RPSLRoute,
    'route-set': RPSLRouteSet,
    'route6': RPSLRoute6,
    'rtr-set': RPSLRtrSet,
}

RPKI_RELEVANT_OBJECT_CLASSES = [
    rpsl_object.rpsl_object_class
    for rpsl_object in OBJECT_CLASS_MAPPING.values()
    if rpsl_object.rpki_relevant
]


def lookup_field_names() -> Set[str]:
    """Return all unique names of all lookup keys in all objects, plus 'origin'."""
    names = {'origin'}
    for object_class in OBJECT_CLASS_MAPPING.values():
        names.update([f for f in object_class.lookup_fields if f not in object_class.pk_fields])
    return names



import logging
import os
import re
from tempfile import NamedTemporaryFile
from typing import Optional, Tuple

import gnupg

from irrd.conf import get_setting

logger = logging.getLogger(__name__)
pgp_inline_re = re.compile(r'-----BEGIN PGP SIGNED MESSAGE-----(\n.+)?\n\n((?s:.+))\n-----BEGIN PGP SIGNATURE-----\n',
                           flags=re.MULTILINE)


def get_gpg_instance() -> gnupg.GPG:
    keyring = get_setting('auth.gnupg_keyring')
    if not os.path.exists(keyring):
        os.mkdir(keyring)
    return gnupg.GPG(gnupghome=keyring)


def validate_pgp_signature(message: str, detached_signature: Optional[str]=None) -> Tuple[Optional[str], Optional[str]]:
    """
    Verify a PGP signature in a message.

    If there is a valid signature, returns a tuple of an optional signed message
    part, and the PGP fingerprint, or None,None if there was no (valid) signature.
    The signed message part is relevant for inline signing, where only part of
    the message may be signed. If it is None, the entire message was signed.

    If detached_signature is set, it is expected to contain a PGP signature block
    that was used to sign message (for PGP/MIME signatures).
    For PGP/MIME, note that message should include the entire text/plain part
    of the signed message, including content-type headers.

    If there is a single PGP inline signed message in message, this message
    will be validated, and the signed part of the message is returned.
    If there are multiple inline PGP signed messages, this function returns
    None,None.

    Note that PGP validation is dependent on the PGP key already being in the
    keychain contained in the auth.gnupg_keyring setting. This is usually done by
    importing a key-cert, which will add the certificate to the keychain during
    validation, in RPSLKeyCert.clean().
    """
    gpg = get_gpg_instance()

    new_message = None
    if detached_signature:
        with NamedTemporaryFile() as data_file:
            data_file.write(message.encode(gpg.encoding))
            data_file.flush()
            result = gpg.verify(detached_signature, data_filename=data_file.name)

    elif message.count('BEGIN PGP SIGNED MESSAGE') == 1:
        result = gpg.verify(message)
        match = pgp_inline_re.search(message.replace('\r\n', '\n').replace('\r', '\n'))
        if not match:  # pragma: no cover
            msg = f'message contained an inline PGP signature, but regular expression failed to extract body: {message}'
            logger.info(msg)
            return None, None
        new_message = match.group(2) + '\n'

    else:
        return None, None

    log_message = result.stderr.replace('\n', ' -- ').replace('gpg:                ', '')
    logger.info(f'checked PGP signature, response: {log_message}')
    if result.valid and result.key_status is None:
        logger.info(f'Found valid PGP signature, fingerprint {result.fingerprint}')
        return new_message, result.fingerprint
    return None, None



import logging
import socket
from typing import List, Tuple, Optional

logger = logging.getLogger(__name__)


class WhoisQueryError(ValueError):
    pass


def whois_query(host: str, port: int, query: str, end_markings: List[str]=None) -> str:
    """
    Perform a query on a whois server, connecting to the specified host and port.

    Will continue to read until no more data can be read, until no more data has
    been sent for 5 seconds, or until an optional end_marking is encountered.
    The end marking could be e.g. 'END NTTCOM' in case of an NRTM stream.
    """
    query = query.strip() + '\n'
    logger.debug(f'Running whois query {query.strip()} on {host} port {port}')
    if end_markings:
        end_markings_bytes = [mark.encode('utf-8') for mark in end_markings]
    else:
        end_markings_bytes = []

    s = socket.create_connection((host, port), timeout=5)
    s.sendall(query.encode('utf-8'))

    buffer = b''
    while not any([end_marking in buffer for end_marking in end_markings_bytes]):
        try:
            data = s.recv(1024*1024)
        except socket.timeout:
            break
        if not data:
            break
        buffer += data
    s.close()

    return buffer.decode('utf-8', errors='backslashreplace')


def whois_query_irrd(host: str, port: int, query: str) -> Optional[str]:
    """
    Perform a whois query, expecting an IRRD-style output format.

    This is a variant of whois_query(), as it uses the additional metadata
    provided by the IRRD output format to know when the full response has
    been received, and whether the full response has been received.
    """
    query = query.strip() + '\n'
    logger.debug(f'Running IRRD whois query {query.strip()} on {host} port {port}')

    s = socket.create_connection((host, port), timeout=5)
    s.sendall(query.encode('utf-8'))

    buffer = b''
    expected_data_length = None
    data_offset = None
    error = None

    while True:
        try:
            data = s.recv(1024)
        except (socket.timeout, ConnectionError):
            break
        if not data:
            break
        buffer += data
        if not expected_data_length:
            decoded_buffer = buffer.decode('utf-8', errors='backslashreplace')
            if '\n' in decoded_buffer:
                length_line = decoded_buffer.splitlines()[0]
                if length_line.startswith('A'):
                    expected_data_length = int(length_line[1:])
                    data_offset = len(length_line) + 1
                elif length_line in ['C', 'D']:
                    break
                elif length_line.startswith('F'):
                    error = length_line[2:]
                    break
        if expected_data_length and data_offset and len(buffer) > (expected_data_length + data_offset):
            break
    s.close()

    if error:
        raise WhoisQueryError(error)
    if not expected_data_length and buffer in [b'C\n', b'D\n']:
        return None
    if not expected_data_length or not data_offset:
        raise ValueError(f'Data receiving ended without a valid IRRD-format response, query {query},'
                         f'received: {buffer.decode("ascii", "ignore")}')
    if len(buffer) < (expected_data_length + data_offset):
        raise ValueError(f'Unable to receive all expected {expected_data_length} bytes')

    return buffer[data_offset:expected_data_length+data_offset-1].decode('utf-8', errors='backslashreplace')


def whois_query_source_status(host: str, port: int, source: str) -> Tuple[Optional[bool], int, int, Optional[int]]:
    """
    Query the status of a particular source against an NRTM server,
    which supports IRRD-style !j queries.

    Will return a tuple with:
    - is this server mirrorable
    - the oldest serial available
    - the newest serial available
    - the serial of the latest export
    """
    remote_status = whois_query_irrd(host, port, f'!j{source}')

    # Fields are: source, mirrorable, serials in journal, optional last export serial
    if not remote_status:
        raise ValueError(f'Source status query on {host}:{port} failed: empty response')
    fields = remote_status.split(':')
    match_source = fields[0].upper()
    if match_source != source.upper():
        raise ValueError(f'Received invalid source {match_source}, expecting {source}, in status: {remote_status}')

    mirrorable_choices = {'Y': True, 'N': False}
    mirrorable = mirrorable_choices.get(fields[1].upper())

    serial_oldest, serial_newest = fields[2].split('-')
    export_serial: Optional[int]
    try:
        export_serial = int(fields[3])
    except IndexError:
        export_serial = None

    return mirrorable, int(serial_oldest), int(serial_newest), export_serial



import datetime
import json
import re
from collections import OrderedDict, Counter
from typing import Dict, List, Optional, Tuple, Any, Set

from IPy import IP

from irrd.rpki.status import RPKIStatus
from irrd.rpsl.parser_state import RPSLParserMessages
from irrd.scopefilter.status import ScopeFilterStatus
from irrd.utils.text import splitline_unicodesafe
from .fields import RPSLTextField
from ..conf import get_setting

RPSL_ATTRIBUTE_TEXT_WIDTH = 16
TypeRPSLObjectData = List[Tuple[str, str, List[str]]]


class RPSLObjectMeta(type):
    """
    Meta class for RPSLObject (and all subclasses) for performance enhancement.

    As RPSLObject is instantiated once per object parsed, __init__ should be
    kept as small as possible. This metaclass pre-calculates some derived data
    from the fields defined by a subclass of RPSLObject, for optimised parsing speed.
    """
    def __init__(cls, name, bases, clsdict):  # noqa: N805
        super().__init__(name, bases, clsdict)
        fields = clsdict.get('fields')
        if fields:
            cls.rpsl_object_class = list(fields.keys())[0]
            cls.pk_fields = [field[0] for field in fields.items() if field[1].primary_key]
            cls.lookup_fields = [field[0] for field in fields.items() if field[1].lookup_key]
            cls.attrs_allowed = [field[0] for field in fields.items()]
            cls.attrs_required = [field[0] for field in fields.items() if not field[1].optional]
            cls.attrs_multiple = [field[0] for field in fields.items() if field[1].multiple]
            cls.referring_strong_fields = [(field[0], field[1].referring) for field in fields.items() if hasattr(field[1], 'referring') and getattr(field[1], 'strong')]


class RPSLObject(metaclass=RPSLObjectMeta):
    """
    Base class for RPSL objects.

    To clean an RPSL object in string form, the best option is not to instance
    this or a subclass, but instead call rpsl_object_from_text() which
    automatically derives the correct class.

    This class should not be instanced directly - instead subclasses should be
    made for each RPSL type with the appropriate fields defined. Note that any
    subclasses should also be added to OBJECT_CLASS_MAPPING.
    """
    fields: Dict[str, RPSLTextField] = OrderedDict()
    rpsl_object_class: str
    pk_fields: List[str] = []
    attrs_allowed: List[str] = []
    attrs_required: List[str] = []
    attrs_multiple: List[str] = []
    ip_first: IP = None
    ip_last: IP = None
    asn_first: Optional[int] = None
    asn_last: Optional[int] = None
    prefix: IP = None  # Note: not saved to DB, used for performance
    prefix_length: Optional[int] = None
    rpki_status: RPKIStatus = RPKIStatus.not_found
    scopefilter_status: ScopeFilterStatus = ScopeFilterStatus.in_scope
    default_source: Optional[str] = None  # noqa: E704 (flake8 bug)
    # Whether this object has a relation to RPKI ROA data, and therefore RPKI
    # checks should be performed in certain scenarios. Enabled for route/route6.
    rpki_relevant = False
    # Fields whose values are discarded during parsing
    discarded_fields: List[str] = []

    _re_attr_name = re.compile(r'^[a-z0-9_-]+$')

    def __init__(self, from_text: Optional[str]=None, strict_validation=True, default_source=None) -> None:
        """
        Create a new RPSL object, optionally instantiated from a string.

        Optionally, you can set/unset strict validation. This means all
        attribute values are validated, and attribute presence/absence is
        verified. Non-strict validation is limited to primary and lookup
        keys.

        If you set default_source, and strict_validation is False, the
        parser will accept objects without a source attribute, and treat
        them as if their source was default_source.
        """
        self.messages = RPSLParserMessages()
        self._object_data: TypeRPSLObjectData = []
        self.strict_validation = strict_validation
        if default_source:
            self.default_source = default_source.strip().upper()

        if from_text:
            self._extract_attributes_values(from_text)
            self._validate_object()

    def pk(self) -> str:
        """Get the primary key value of an RPSL object. The PK is always converted to uppercase."""
        if len(self.pk_fields) == 1:
            return self.parsed_data.get(self.pk_fields[0], "").upper()
        composite_values = []
        for field in self.pk_fields:
            composite_values.append(self.parsed_data.get(field, ""))
        return ''.join(composite_values).upper()

    def source(self) -> str:
        """Shortcut to retrieve object source"""
        try:
            return self.parsed_data['source']
        except KeyError:
            raise ValueError('RPSL object has no known source')

    def ip_version(self) -> Optional[int]:
        """
        Get the IP version to which this object relates, or None for
        e.g. person or as-block objects.
        """
        if self.ip_first:
            return self.ip_first.version()
        return None

    def referred_strong_objects(self) -> List[Tuple[str, List, List]]:
        """
        Get all objects that this object refers to (e.g. an admin-c attribute
        on this object, that refers to person/role) along with the data this
        object has for that reference. This information can be used to check
        whether all references from an object are valid.
        Only references which have strong=True are returned, weak references
        are not returned as they should not be included in reference validation.

        Returns a list of tuples, which each tuple having:
        - field name on this object (e.g. 'admin-c')
        - RPSL object class names of objects referred (e.g. ['role', 'person']
        - Values this object has for that field (e.g. ['A-RIPE', 'B-RIPE']
        """
        result = []
        for field_name, referred_objects in self.referring_strong_fields:  # type: ignore
            data = self.parsed_data.get(field_name)
            if not data:
                continue
            result.append((field_name, referred_objects, data))
        return result

    def references_strong_inbound(self) -> Set[str]:
        """
        Get a set of field names under which other objects refer to
        this object. E.g. for a person object, this would typically
        return {'zone-c', 'admin-c', 'tech-c'}.
        """
        result = set()
        from irrd.rpsl.rpsl_objects import OBJECT_CLASS_MAPPING
        for rpsl_object in OBJECT_CLASS_MAPPING.values():
            for field_name, field in rpsl_object.fields.items():
                if self.rpsl_object_class in getattr(field, 'referring', []) and getattr(field, 'strong'):
                    result.add(field_name)
        return result

    def render_rpsl_text(self, last_modified: datetime.datetime=None) -> str:
        """
        Render the RPSL object as an RPSL string.
        If last_modified is provided, removes existing last-modified:
        attributes and adds a new one with that timestamp, if self.source()
        is authoritative.
        """
        output = ""
        authoritative = get_setting(f'sources.{self.source()}.authoritative')
        for attr, value, continuation_chars in self._object_data:
            if authoritative and last_modified and attr == 'last-modified':
                continue
            attr_display = f'{attr}:'.ljust(RPSL_ATTRIBUTE_TEXT_WIDTH)
            value_lines = list(splitline_unicodesafe(value))
            if not value_lines:
                output += f'{attr}:\n'
            for idx, line in enumerate(value_lines):
                if idx == 0:
                    output += attr_display + line
                else:
                    continuation_char = continuation_chars[idx - 1]
                    # Override the continuation char for empty lines #298
                    if not line:
                        continuation_char = '+'
                    output += continuation_char + (RPSL_ATTRIBUTE_TEXT_WIDTH - 1) * ' ' + line
                output += '\n'
        if authoritative and last_modified:
            output += 'last-modified:'.ljust(RPSL_ATTRIBUTE_TEXT_WIDTH)
            output += last_modified.replace(microsecond=0).isoformat().replace('+00:00', 'Z')
            output += '\n'
        return output

    def generate_template(self):
        """Generate a template in text form of the main attributes of all fields."""
        template = ""
        max_name_width = max(len(k) for k in self.fields.keys())
        for name, field in self.fields.items():
            mandatory = '[optional] ' if field.optional else '[mandatory]'
            single = '[multiple]' if field.multiple else '[single]  '
            metadata = []
            if field.primary_key and field.lookup_key:
                metadata.append('primary/look-up key')
            elif field.primary_key:
                metadata.append('primary key')
            elif field.lookup_key:
                metadata.append('look-up key')
            if getattr(field, 'referring', []):
                reference_type = 'strong' if getattr(field, 'strong') else 'weak'
                metadata.append(f'{reference_type} references ' + '/'.join(field.referring))
            metadata_str = ', '.join(metadata)
            name_padding = (max_name_width - len(name)) * ' '
            template += f'{name}: {name_padding}  {mandatory}  {single}  [{metadata_str}]\n'
        return template

    def clean(self) -> bool:
        return True

    def _extract_attributes_values(self, text: str) -> None:
        """
        Extract all attributes and associated values from the input string.

        This is mostly straight forward, except for the tricky feature of line
        continuation. An attribute's value can be continued on the next lines,
        which is distinct from an attribute occurring multiple times.

        The parse result is internally stored in self._object_data. This is a
        list of 3-tuples, where each tuple contains the attribute name,
        attribute value, and the continuation characters. The continuation
        characters are needed to reconstruct the original object into a string.
        """
        continuation_chars = (' ', '+', '\t')
        current_attr = None
        current_value = ""
        current_continuation_chars: List[str] = []

        for line_no, line in enumerate(splitline_unicodesafe(text.strip())):
            if not line:
                self.messages.error(f'Line {line_no+1}: encountered empty line in the middle of object: [{line}]')
                return

            if not line.startswith(continuation_chars):
                if current_attr and current_attr not in self.discarded_fields:
                    # Encountering a new attribute requires saving the previous attribute data first, if any,
                    # which can't be done earlier as line continuation means we can't know earlier whether
                    # the attribute is finished.
                    self._object_data.append((current_attr, current_value, current_continuation_chars))

                if ':' not in line:
                    self.messages.error(f'Line {line_no+1}: line is neither continuation nor valid attribute [{line}]')
                    return
                current_attr, current_value = line.split(':', maxsplit=1)
                current_attr = current_attr.lower()
                current_value = current_value.strip()
                current_continuation_chars = []

                if current_attr not in self.attrs_allowed and not self._re_attr_name.match(current_attr):
                    self.messages.error(f'Line {line_no+1}: encountered malformed attribute name: [{current_attr}]')
                    return
            else:
                # Whitespace between the continuation character and the start of the data is not significant.
                current_value += '\n' + line[1:].strip()
                current_continuation_chars += line[0]
        if current_attr and current_attr not in self.discarded_fields:
            self._object_data.append((current_attr, current_value, current_continuation_chars))

    def _validate_object(self) -> None:
        """
        Validate an object. The strictness depends on self.strict_validation
        (see the docstring for __init__).
        """
        self.parsed_data: Dict[str, Any[str, List]] = {}
        if not self.messages.errors():
            self._validate_attribute_counts()
        self._parse_attribute_data(allow_invalid_metadata=bool(self.messages.errors()))

        if self.strict_validation and not self.messages.errors():
            self.clean()

    def _validate_attribute_counts(self) -> None:
        """
        Validate the number of times each attribute occurs.

        The expected counts (0, 1, or >=1) are derived indirectly
        from the field data. In non-strict mode, only validate
        presence of all PK attributes.
        """
        attrs_present = Counter([attr[0] for attr in self._object_data])

        if self.strict_validation:
            for attr_name, count in attrs_present.items():
                if attr_name not in self.attrs_allowed:
                    self.messages.error(f'Unrecognised attribute {attr_name} on object {self.rpsl_object_class}')
                if count > 1 and attr_name not in self.attrs_multiple:
                    self.messages.error(
                        f'Attribute "{attr_name}" on object {self.rpsl_object_class} occurs multiple times, but is '
                        f'only allowed once')
            for attr_required in self.attrs_required:
                if attr_required not in attrs_present:
                    self.messages.error(
                        f'Mandatory attribute "{attr_required}" on object {self.rpsl_object_class} is missing'
                    )
        else:
            required_fields = self.pk_fields
            if not self.default_source:
                required_fields = required_fields + ['source']
            for attr_pk in required_fields:
                if attr_pk not in attrs_present:
                    self.messages.error(
                        f'Primary key attribute "{attr_pk}" on object {self.rpsl_object_class} is missing'
                    )

    def _parse_attribute_data(self, allow_invalid_metadata=False) -> None:
        """
        Clean the data stored in attributes.

        If self.strict_validation is not set, only checks primary and lookup keys,
        as they need to be indexed. All parsed values (e.g. without comments) are
        stored in self.parsed_data - stored in upper case unless a field is marked
        case sensitive.

        If allow_invald_metadata is set, the parser will accept invalid metadata
        being stored, but still make a best effort to extract data. This should
        only be used if it is already known the object is invalid, and will
        never be stored.
        """
        for idx, (attr_name, value, continuation_chars) in enumerate(self._object_data):
            field = self.fields.get(attr_name)
            if field:
                normalised_value = self._normalise_rpsl_value(value)

                # We always parse all fields, but only care about errors if we're running
                # in strict validation mode, if the field is primary or lookup, or if it's
                # the source field. In all other cases, the field parsing is best effort.
                # In all these other cases we pass a new parser messages object to the
                # field parser, so that we basically discard any errors.
                raise_errors = self.strict_validation or field.primary_key or field.lookup_key or attr_name == 'source'
                field_messages = self.messages if raise_errors else RPSLParserMessages()
                parsed_value = field.parse(normalised_value, field_messages, self.strict_validation)

                if parsed_value:
                    parsed_value_str = parsed_value.value
                    if parsed_value_str != normalised_value:
                        # Note: this replacement can be incomplete: if the normalised value is not contained in the
                        # parsed value as single string, the replacement will not occur. This is not a great concern,
                        # as this is purely cosmetic, and self.parsed_data will have the correct normalised value.
                        new_value = value.replace(normalised_value, parsed_value_str)
                        self._object_data[idx] = attr_name, new_value, continuation_chars
                    values_list = parsed_value.values_list
                    if values_list:
                        if not field.keep_case:
                            values_list = list(map(str.upper, values_list))
                        if attr_name in self.parsed_data:
                            self.parsed_data[attr_name] += values_list
                        else:
                            self.parsed_data[attr_name] = values_list
                    else:
                        if not field.keep_case:
                            parsed_value_str = parsed_value_str.upper()
                        if field.multiple:
                            if attr_name in self.parsed_data:
                                self.parsed_data[attr_name].append(parsed_value_str)
                            else:
                                self.parsed_data[attr_name] = [parsed_value_str]
                        else:
                            if attr_name in self.parsed_data:
                                self.parsed_data[attr_name] = '\n' + parsed_value_str
                            else:
                                self.parsed_data[attr_name] = parsed_value_str

                    # Some fields provide additional metadata about the resources to
                    # which this object pertains.
                    if field.primary_key or field.lookup_key:
                        for attr in 'ip_first', 'ip_last', 'asn_first', 'asn_last', 'prefix', 'prefix_length':
                            attr_value = getattr(parsed_value, attr, None)
                            if attr_value:
                                existing_attr_value = getattr(self, attr, None)
                                if existing_attr_value and not allow_invalid_metadata:  # pragma: no cover
                                    raise ValueError(f'Parsing of {parsed_value.value} reads {attr_value} for {attr},'
                                                     f'but value {existing_attr_value} is already set.')
                                setattr(self, attr, attr_value)

        if 'source' not in self.parsed_data and self.default_source:
            self.parsed_data['source'] = self.default_source

    def _normalise_rpsl_value(self, value: str) -> str:
        """
        Normalise an RPSL attribute value to its significant parts
        in a consistent format.

        For example, the following is valid in RPSL:

            inetnum: 192.0.2.0 # comment1
            +- # comment 2
            +192.0.2.1 # comment 3
            + # comment 4

        This value will be normalised by this method to:
            192.0.2.0 - 192.0.2.1
        to be used for further validation and extraction of primary keys.
        """
        normalized_lines = []
        # The shortcuts below are functionally inconsequential, but significantly improve performance,
        # as most values are single line without comments, and this method is called extremely often.
        if '\n' not in value:
            if '#' in value:
                return value.split('#')[0].strip()
            return value.strip()
        for line in splitline_unicodesafe(value):
            parsed_line = line.split('#')[0].strip('\n\t, ')
            if parsed_line:
                normalized_lines.append(parsed_line)
        return ','.join(normalized_lines)

    def _update_attribute_value(self, attribute, new_values):
        """
        Update the value of an attribute in the internal state and in
        parsed_data.

        This is used for key-cert objects, where e.g. owner lines are
        derived from other data in the object.

        All existing occurences of the attribute are removed, new items
        are always inserted at line 2 of the object.
        """
        if isinstance(new_values, str):
            new_values = [new_values]
        self.parsed_data[attribute] = '\n'.join(new_values)

        self._object_data = list(filter(lambda a: a[0] != attribute, self._object_data))
        insert_idx = 1
        for new_value in new_values:
            self._object_data.insert(insert_idx, (attribute, new_value, []))
            insert_idx += 1

    def __repr__(self):
        source = self.parsed_data.get('source', '')
        return f'{self.rpsl_object_class}/{self.pk()}/{source}'

    def __key(self):
        return self.rpsl_object_class, self.pk(), json.dumps(self.parsed_data, sort_keys=True)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(self, type(other)) and self.__key() == other.__key()


class UnknownRPSLObjectClassException(Exception):
    def __init__(self, message: str, rpsl_object_class: str) -> None:
        self.message = message
        self.rpsl_object_class = rpsl_object_class

    def __str__(self):
        return self.message



import logging
import multiprocessing as mp
import os
import signal
import socket
import socketserver
import threading
import time

from IPy import IP
from setproctitle import setproctitle

from irrd.conf import get_setting, get_configuration
from irrd.server.access_check import is_client_permitted
from irrd.server.whois.query_parser import WhoisQueryParser
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.preload import Preloader
from irrd.utils.process_support import memory_trim

logger = logging.getLogger(__name__)
mp.allow_connection_pickling()


# Covered by integration tests
def start_whois_server():  # pragma: no cover
    """
    Start the whois server, listening forever.
    This function does not return, except after SIGTERM is received.
    """
    setproctitle('irrd-whois-server-listener')
    address = (get_setting('server.whois.interface'), get_setting('server.whois.port'))
    logger.info(f'Starting whois server on TCP {address}')
    server = WhoisTCPServer(
        server_address=address,
    )

    # When this process receives SIGTERM, shut down the server cleanly.
    def sigterm_handler(signum, frame):
        nonlocal server

        def shutdown(server):
            logging.info('Whois server shutting down')
            server.shutdown()
            server.server_close()
        # Shutdown must be called from a thread to prevent blocking.
        threading.Thread(target=shutdown, args=(server,)).start()
    signal.signal(signal.SIGTERM, sigterm_handler)

    def sighup_handler(signum, frame):
        nonlocal server

        def sighup(server):
            get_configuration().reload()
            server.sighup_workers()
        # sighup must be called from a thread to prevent blocking.
        threading.Thread(target=sighup, args=(server,)).start()
    signal.signal(signal.SIGHUP, sighup_handler)

    server.serve_forever()


class WhoisTCPServer(socketserver.TCPServer):  # pragma: no cover
    """
    Server for whois queries.

    Starts a number of worker processes that handle the client connections.
    Whenever a client is connected, the connection is pushed onto a queue,
    from which a worker picks it up. The workers are responsible for the
    connection from then on.
    """
    allow_reuse_address = True
    request_queue_size = 50

    def __init__(self, server_address, bind_and_activate=True):  # noqa: N803
        self.address_family = socket.AF_INET6 if IP(server_address[0]).version() == 6 else socket.AF_INET
        super().__init__(server_address, None, bind_and_activate)

        self.connection_queue = mp.Queue()
        self.workers = []
        for i in range(int(get_setting('server.whois.max_connections'))):
            worker = WhoisWorker(self.connection_queue)
            worker.start()
            self.workers.append(worker)

    def process_request(self, request, client_address):
        """Push the client connection onto the queue for further handling."""
        self.connection_queue.put((request, client_address))

    def handle_error(self, request, client_address):
        logger.error(f'Error while handling request from {client_address}', exc_info=True)

    def sighup_workers(self):
        for worker in self.workers:
            os.kill(worker.pid, signal.SIGHUP)

    def shutdown(self):
        """
        Shut down the server, by killing all child processes,
        and then deferring to built-in TCPServer shutdown.
        """
        for worker in self.workers:
            try:
                worker.terminate()
                worker.join()
            except Exception:  # pragma: no cover
                pass
        return super().shutdown()


class WhoisWorker(mp.Process, socketserver.StreamRequestHandler):
    """
    A whois worker is a process that handles whois client connections,
    which are retrieved from a queue. After handling a connection,
    the process waits for the next connection from the queue.s
    """
    def __init__(self, connection_queue, *args, **kwargs):
        self.connection_queue = connection_queue
        # Note that StreamRequestHandler.__init__ is not called - the
        # input for that is not available, as it's retrieved from the queue.
        super().__init__(*args, **kwargs)

    def run(self, keep_running=True) -> None:
        """
        Whois worker run loop.
        This method does not return, except if it failed to initialise a preloader,
        or if keep_running is set, after the first request is handled. The latter
        is used in the tests.
        """
        # Disable the special sigterm_handler defined in start_whois_server()
        # (signal handlers are inherited)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

        try:
            self.preloader = Preloader()
            self.database_handler = DatabaseHandler(readonly=True)
        except Exception as e:
            logger.error(f'Whois worker failed to initialise preloader or database,'
                         f'unable to start, traceback follows: {e}', exc_info=e)
            return

        while True:
            try:
                setproctitle('irrd-whois-worker')
                self.request, self.client_address = self.connection_queue.get()
                self.setup()
                self.handle_connection()
                self.finish()
                self.close_request()
                memory_trim()
            except Exception as e:
                try:
                    self.close_request()
                except Exception:  # pragma: no cover
                    pass
                logger.error(f'Failed to handle whois connection, traceback follows: {e}',
                             exc_info=e)
            if not keep_running:
                break

    def close_request(self):
        # Close the connection in the same way normally done by TCPServer
        try:
            # explicitly shutdown.  socket.close() merely releases
            # the socket and waits for GC to perform the actual close.
            self.request.shutdown(socket.SHUT_WR)
        except OSError:  # pragma: no cover
            pass  # some platforms may raise ENOTCONN here
        self.request.close()

    def handle_connection(self):
        """
        Handle an individual whois client connection.
        When this method returns, the connection is closed.
        """
        client_ip = self.client_address[0]
        self.client_str = client_ip + ':' + str(self.client_address[1])
        setproctitle(f'irrd-whois-worker-{self.client_str}')

        if not self.is_client_permitted(client_ip):
            self.wfile.write(b'%% Access denied')
            return

        self.query_parser = WhoisQueryParser(client_ip, self.client_str, self.preloader,
                                             self.database_handler)

        data = True
        while data:
            timer = threading.Timer(self.query_parser.timeout, self.close_request)
            timer.start()
            data = self.rfile.readline()
            timer.cancel()

            query = data.decode('utf-8', errors='backslashreplace').strip()
            if not query:
                continue

            logger.debug(f'{self.client_str}: processing query: {query}')

            if not self.handle_query(query):
                return

    def handle_query(self, query: str) -> bool:
        """
        Handle an individual query.
        Returns False when the connection should be closed,
        True when more queries should be read.
        """
        start_time = time.perf_counter()
        if query.upper() == '!Q':
            logger.debug(f'{self.client_str}: closed connection per request')
            return False

        response = self.query_parser.handle_query(query)
        response_bytes = response.generate_response().encode('utf-8')
        try:
            self.wfile.write(response_bytes)
        except OSError:
            return False

        elapsed = time.perf_counter() - start_time
        logger.info(f'{self.client_str}: sent answer to query, elapsed {elapsed:.9f}s, '
                    f'{len(response_bytes)} bytes: {query}')

        if not self.query_parser.multiple_command_mode:
            logger.debug(f'{self.client_str}: auto-closed connection')
            return False
        return True

    def is_client_permitted(self, ip: str) -> bool:
        """
        Check whether a client is permitted.
        """
        return is_client_permitted(ip, 'server.whois.access_list', default_deny=False)
